@@ -127,8 +127,9 @@ def socket_send(sock, flood=False):
             incoming.put(("sending",))
             while True:
                 octets = outgoing.get()
-                if octets == None:
-                    break
+                if octets is None:
+                    debug("Sending Thread: Requested quit")
+                    return True
 
                 debug("->", repr(octets.decode("utf-8", "replace")))
                 s.write(octets)
@@ -136,11 +137,13 @@ def socket_send(sock, flood=False):
                 if not flood:
                     # TODO: Allow two or three burst lines
                     time.sleep(1)
+        debug("Sending Thread: No more data")
+        return False
 
     try: sending(sock, flood)
     except Exception as err:
         # Usually BrokenPipeError
-        ...
+        debug("Sending Thread: Error:", err)
     incoming.put(("disco_sending",))
 
 class SaxoConnectionError(Exception):
@@ -155,7 +158,9 @@ class Saxo(object):
         self.discotimer = None
         self.receiving = False
         self.sending = False
-        self.user_reconnection = False
+        self.receiving_thread = None
+        self.sending_thread = None
+        self.reconnecting = False
         self.links = {}
 
         self.environment_cache = os.environ.copy()
@@ -235,8 +240,13 @@ class Saxo(object):
            raise SaxoConnectionError(str(err))
 
         self.first = True
-        common.thread(socket_receive, self.sock)
-        common.thread(socket_send, self.sock, "flood" in self.opt["client"])
+        # TODO: Reset other state? e.g. self.address
+
+        receiving = (socket_receive, self.sock)
+        self.receiving_thread = common.thread(*receiving)
+
+        sending = (socket_send, self.sock, "flood" in self.opt["client"])
+        self.sending_thread = common.thread(*sending)
 
     def connect_sock(self):
         host = self.opt["server"]["host"]
@@ -255,9 +265,34 @@ class Saxo(object):
         self.sock.connect((host, port))
 
     def disconnect(self):
-        outgoing.put(None) # Closes the send thread
-        self.sock.shutdown(socket.SHUT_RDWR)
-        self.sock.close()
+        # NOTE: *Do not* close the sending thread gracefully
+        # The socket shutdown code below will be invoked before the queue
+        # item reaches the socket thread. That means the *next* thread will
+        # pick it up, because it'll quit with a pipe error anyway
+
+        # Close the socket forcefully
+        try: self.sock.shutdown(socket.SHUT_RDWR)
+        except: ...
+
+        try: self.sock.close()
+        except: ...
+
+        self.cancel_discotimer()
+
+    def socket_threads_active(self):
+        receiving = self.receiving_thread.is_alive()
+        sending = self.sending_thread.is_alive()
+        debug("RECV, SEND:", receiving, sending)
+        return receiving or sending
+
+    def cancel_discotimer(self):
+        if self.discotimer is not None:
+            try:
+                self.discotimer.cancel()
+                self.discotimer = None
+                debug("Cancelled the disco timer")
+            except:
+                ...
 
     def handle(self):
         while True:
@@ -265,7 +300,7 @@ class Saxo(object):
             instruction = instruction_args[0]
             args = tuple(instruction_args[1:])
 
-            if instruction != "remote":
+            if instruction not in {"instances", "remote"}:
                 debug("handle:", instruction, args)
 
             if not isinstance(instruction, str):
@@ -284,6 +319,32 @@ class Saxo(object):
     def instruction_address(self, address):
         self.address = address
 
+    def instruction_connect(self, delay=5):
+        # Warning: This quits the bot if the bot is already connected!
+        # Use instruction_reconnect if you want to restart the bot
+
+        # NOTE: The instruction mechanism is synchronous
+        # This means we can't use self.receiving etc. to check connectivity
+        # We rely on self.disconnect() to have done the right thing
+        # The following is a check to make sure that it has
+        if self.socket_threads_active():
+           time.sleep(2)
+           if self.socket_threads_active():
+               # TODO: If the threads are still active, they should be killed
+               # Unfortunately, threads in python can't be killed
+               debug("ERROR! Unable to stop the socket threads")
+               os._exit(1)
+
+        if "flood" not in self.opt["client"]:
+            time.sleep(3)
+
+        try: self.connect()
+        except SaxoConnectionError as err:
+            def connect():
+                incoming.put(("connect", min(30, delay + 1)))
+            t = threading.Timer(delay, connect)
+            t.start()
+
     def instruction_connected(self):
         if ":connected" in self.events:
             irc = ThreadSafeEnvironment(self, ("", "", ""), "", [])
@@ -295,20 +356,17 @@ class Saxo(object):
 
     def instruction_disco_receiving(self):
         self.receiving = False
-        if not self.user_reconnection:
-            if self.sending:
-                outgoing.put(None)
-            else:
-                incoming.put(("reconnect", False))
+        if self.sending:
+            outgoing.put(None)
+        else:
+            incoming.put(("connect",))
 
     def instruction_disco_sending(self):
         self.sending = False
-        if not self.user_reconnection:
-            if not self.receiving:
-                incoming.put(("reconnect", False))
-            else:
-                # TODO: Close the socket?
-                ...
+        if self.receiving:
+            self.disconnect()
+        else:
+            incoming.put(("connect",))
 
     def instruction_instances(self):
         our_pid = os.getpid()
@@ -347,6 +405,9 @@ class Saxo(object):
     def instruction_msg(self, destination, text):
         self.send("PRIVMSG", destination, text)
 
+    def instruction_noop(self):
+        ...
+
     def instruction_part(self, channel):
         # NOTE: .leave can still be followed by .part
         # The PART will just fail on the server
@@ -362,17 +423,20 @@ class Saxo(object):
         self.send("PART", channel)
 
     def instruction_ping(self):
-        # Don't ping if not connected, or connected less than a minute ago
-        if not self.receiving:
-            return
         now = time.time()
+
+        # Don't ping if connected less than a minute ago
         if self.receiving > (now - 60):
+            debug("Won't start discotimer, because we recently reconnected")
             return
 
         self.send("PING", self.opt["client"]["nick"])
 
         def reconnect():
             incoming.put(("reconnect",))
+
+        # Make sure the timer is shorter than the pingloop!
+        # Default pingloop period is 180
         self.discotimer = threading.Timer(30, reconnect)
         self.discotimer.start()
 
@@ -396,28 +460,14 @@ class Saxo(object):
         # TODO: Unit test for :connected event
         incoming.put(("connected",))
 
-    def instruction_reconnect(self, close=True, wait=3):
-        if close:
-            self.user_reconnection = True
-            # Never call this from a thread, otherwise this can give an OSError
+    def instruction_reconnect(self, wait=3):
+        # disco_* will automatically reconnect
+        if self.receiving or self.receiving_thread.is_alive():
             self.disconnect()
-
-        flood = "flood" in self.opt["client"]
-        if not flood:
-            time.sleep(wait)
-        for attempt in range(7):
-            if not flood:
-                if self.receiving or self.sending:
-                    time.sleep(1)
-                # TODO: If the threads are still active, they should be killed
-                # Unfortunately, threads in python can't be killed
-
-        if close:
-            self.user_reconnection = False
-
-        try: self.connect()
-        except SaxoConnectionError as err:
-            incoming.put(("reconnect", close, 20))
+        elif self.sending or self.sending_thread.is_alive():
+            outgoing.put(None)
+        else:
+            debug("Reconnect requested during reconnect")
 
     def instruction_reload(self, destination=None):
         before = time.time()
@@ -457,13 +507,7 @@ class Saxo(object):
                 self.command(prefix, sender, cmd, arg)
 
         elif command == "PONG":
-            if self.discotimer is not None:
-                try:
-                    self.discotimer.cancel()
-                    self.discotimer = None
-                    debug("Cancelled the disco timer")
-                except:
-                    ...
+            self.cancel_discotimer()
 
         args = (self, prefix, command, parameters)
         irc = ThreadSafeEnvironment(*args)
