@@ -20,14 +20,28 @@ incoming = queue.Queue()
 def start(base, client):
     database_filename = os.path.join(base, "database.sqlite3")
     with sqlite.Database(database_filename) as db:
+        if not "saxo_instances" in db:
+            db["saxo_instances"].create(
+                ("pid", int))
+
+        # Remove any values from previous instances
+        for row in db["saxo_instances"]:
+            del db["saxo_instances"][row]
+        # Add our value
+        db["saxo_instances"].insert((os.getpid(),))
+
         if not "saxo_periodic" in db:
             db["saxo_periodic"].create(
                 ("period", int),
                 ("command", bytes),
                 ("args", bytes))
+        else:
+            for row in db["saxo_periodic"]:
+                del db["saxo_periodic"][row]
 
-            # TODO: Or "check_connection"
-            db["saxo_periodic"].insert((180, b"ping", b""))
+        # TODO: Or "check_connection" instead of "ping"
+        db["saxo_periodic"].insert((180, b"ping", b""))
+        db["saxo_periodic"].insert((20, b"instances", b""))
 
         if not "saxo_schedule" in db:
             db["saxo_schedule"].create(
